/**********************************************************************/
/*      Licensed Materials - Property of IBM                          */
/*      "Restricted Materials of IBM"                                 */
/*      5748-T15                                                      */
/*      Copyright IBM Corp. 2016, 2019                                */
/*      All Rights Reserved                                           */
/*      US Government Users Restricted Rights - Use, duplication      */
/*      or disclosure restricted by GSA ADP Schedule Contract         */
/*      with IBM Corp.                                                */
/**********************************************************************/
/*                                                                    */
/* FUNCTION NAME.....connserver                                       */
/* PURPOSE...........connserver driver for High Speed Connector       */
/* FUNCTION TYPE.....Linux                                            */
/*                                                                    */
/* INPUT.......none                                                   */
/*                                                                    */
/*                                                                    */
/* Options:    -i SERVER_ADDRESS                                      */
/*                  Where SERVER_ADDRESS is a valid IPv4 address.     */
/*                  Default value is 0.0.0.0                          */
/*             -p PORT                                                */
/*                  Where PORT is the port for the server to listen on*/
/*                  0 <= PORT <= 65535                                */
/*                  This option is REQUIRED.                          */
/*             -d RESPONSE_DELAY                                      */
/*                  Must be an integer between 0 and 600000, inclusive*/
/*                  Default value is 0.                               */
/*             -s SEND_BUFFER_SIZE                                    */
/*                  Size of buffer in bytes in which messages are     */
/*                      sent.  Only used if RESP_REQ=1                */
/*                  Default value is 8192.                            */
/*                  Must be an integer between 512 and 1048576,       */
/*                      inclusive.                                    */
/*             -r RCV_BUFFER_SIZE                                     */
/*                  Size of buffer in bytes in which messages are     */
/*                      received from the tpf_send_message API.       */
/*                  Default value is 8192.                            */
/*                  Must be an integer between 512 and 1048576,       */
/*                      inclusive.                                    */
/*             -q RESP_REQ                                            */
/*                  Defines whether or not this server will send a    */
/*                      response when a message is received.          */
/*                  Default value is 0.                               */
/*                  Must be either 0 or 1 -- 0 or no, 1 for yes.      */
/*             -m RESP_MSG_SIZE                                       */
/*                  The size, in bytes, of the response message       */
/*                      sent by connserver if RESP_REQ=1.             */
/*                  Default value is 4096.                            */
/*             -t                                                     */
/*                  Specify the t flag to enable SSL. Default is      */
/*                  no SSL.                                           */
/*                                                                    */
/* DESCRIPTION: QHSC is the program that tests the high               */
/*              speed connector API function. It acts as the          */
/*              client. connserver is the server portion that         */
/*              runs on Linux to respond to requests from QHSC.       */
/*                                                                    */
/*                                                                    */
/**********************************************************************/
/*                                                                    */
/* DEPENDENCIES... None.                                              */
/*                                                                    */
/* RESTRICTIONS... None.                                              */
/*                                                                    */
/* SPECIAL CONSIDERATIONS...                                          */
/*                                                                    */
/* INTERFACES:                                                        */
/*                                                                    */
/*                                                                    */
/**********************************************************************/
/*                                                                    */
/* CHANGE ACTIVITY...                                                 */
/*                                                                    */
/* DEFECT       DATE       DESCRIPTION                                */
/*   ----       ----       -----------                                */
/* f26077       03/28/2016 Create High Speed Connector driver         */
/* d26288       12/18/2017 Remove connapi.h                           */
/* d26322       08/24/2018 Added SSL option to existing connserver    */
/* d26418       11/26/2018 TLS configuration and driver changes       */
/* d26480       04/19/2019 Updates to connserver                      */
/**********************************************************************/

#include <unistd.h>                  //close, read, usleep
#include <stdio.h>                   //printf, fflosh
#include <stdlib.h>                  //exit, free, atoi
#include <string.h>                  //memcpy, memset
#include <arpa/inet.h>               //inet_aton, inet_ntoa, htons, sockaddr_in
#include <sys/socket.h>              //socket, bind, listen, accept
#include <pthread.h>                 //pthread_exit, pthread_create
#include <openssl/ssl.h>             //OpenSSL functions                        //d26322 

#define HSC_HEADER_SIZE 32           // HSC header length (from connapi.h)      //d26288
#define MAX_BUFFER_SIZE 1048576      // Maximum send/receive buffer size
#define MIN_BUFFER_SIZE 512          // Minimum send/receive buffer size
#define MAX_DELAY 600000             // Maximum allowed delay
#define MIN_DELAY 0                  // Minimum allowed delay

/* Certificate and key file paths */
#define CERT_FILE "yourCert.pem"                                                //d26418
#define KEY_FILE "yourKey.pem"                                                  //d26418
#define KEY_PASSWD "yourPassword"                                               //d26418

/* Cipher list used */
#define CIPHER_LIST "Your:Cipher:List"                                          //d26418 

//DEFAULT OPTION VALUES
struct sockaddr_in server_address;   //i
int port = -1;                       //p
int response_delay = 0;              //d
int send_buffer_size = 8192;         //s
int receive_buffer_size = 8192;      //r
int response_required = 0;           //q
int response_message_size = 4096;    //m
int SSL_enabled = 0;                 //t                                        //d26322

/* High Speed Connector response structure from connapi.h */
struct hsc_resp {                                                               //d26288
   int length;           /* length of response data, not including this field *///d26288
   unsigned short header_length;                                                //d26288
   short rc;             /* return code                                       *///d26288
   char reserve[24];     /* Reserved for IBM use                              *///d26288
   char data[1];         /* Response data (variable length)                   *///d26288
};                                                                              //d26288

/* Structure to hold the arguments passed into the worker thread */
struct thread_args {                                                            //d26322
   SSL_CTX *ssl_ctx;     /* SSL_CTX is collection of ciphers, certificates etc. */
   int      sock_fd;     /* socket file descriptor */                           //d26322
};                                                                              //d26322

/* Error handling routines for the base program, the socket, and the thread */
void conn_err(char *msg) {
    printf("CONN: %s\n",msg);
    exit(0);
}

/* Error routine for after socket is open */
void conn_err_s(char* msg, int sockfd, SSL_CTX *ctx) {
    if (SSL_enabled)                                                            //d26322
        SSL_CTX_free(ctx);                                                      //d26322 

    printf("CONN: %s, closing socket no. %d\n",msg, sockfd);
    close(sockfd);
    exit(0);
}

/* Error routine for use in thread handler */
void conn_err_t(char *msg, int connsock, SSL *ssl_token) {                      //d26322
    /* Additional cleanup req if SSL connection specified */
    if (ssl_token && SSL_enabled)                                               //d26322   
        SSL_free(ssl_token);                                                    //d26322 

    printf("CONN: %s\n",msg);
    close(connsock);
    pthread_exit(NULL);
}

/* Returns newly created pointer to SSL_CTX which has the context (information) */
/* about the SSL session. Holds mainly default values for SSL structures  later created*/ 
SSL_CTX* SSL_init_server_CTX() {
    const SSL_METHOD *method;     // struct describing SSL methods that implement protocols
    SSL_CTX *ctx;                                                               //d26322

    SSL_library_init();                                                         //d26322 
    OpenSSL_add_all_algorithms();                                               //d26322
    SSL_load_error_strings();                                                   //d26322

    /* Negotiate highest avail TLS/SSL version for the connection */
    method = SSLv23_server_method();                                            //d26322
    if (!method) {                                                              //d26322 
        printf("Could not create method\n");                                    //d26322
        exit(0);                                                                //d26322
    }                                                                           //d26322

    /* Create the new CTX context with the method*/
    ctx = SSL_CTX_new(method);                                                  //d26322 
    if (!ctx) {                                                                 //d26322 
       printf("Error creating SSL context.\n");                                 //d26322
       exit(0);                                                                 //d26322 
    }                                                                           //d26322

    return ctx;                                                                 //d26322
}                                                                               //d26322

/* Setup with the private key and certificate file */
void LoadCertificates(SSL_CTX* ctx, char* cert, char* key) {                    //d26322
    /* Set local certificate */
    if (SSL_CTX_use_certificate_file(ctx, cert, SSL_FILETYPE_PEM) <= 0) {       //d26322
       printf("Error setting up certificate file.\n");                          //d26322
       SSL_CTX_free(ctx);                                                       //d26322
       exit(0);                                                                 //d26322
    }                                                                           //d26322

    /* Load the password for the Private Key */
    SSL_CTX_set_default_passwd_cb_userdata(ctx, KEY_PASSWD);                    //d26322 

    /* Set private key */
    if (SSL_CTX_use_PrivateKey_file(ctx, key, SSL_FILETYPE_PEM) <= 0) {         //d26322
       printf("Error setting up key file.\n");                                  //d26322  
       SSL_CTX_free(ctx);                                                       //d26322
       exit(0);                                                                 //d26322 
    }                                                                           //d26322

    /* Verify private key */
    if (!SSL_CTX_check_private_key(ctx)) {                                      //d26322
       printf("Private key does not match public certificate.\n");              //d26322
       SSL_CTX_free(ctx);                                                       //d26322
       exit(0);                                                                 //d26322
    }                                                                           //d26322
}                                                                               //d26322

/* Wrapper function to use regular read function for non-SSL connections */
/* or SSL_read if SSL is enabled.                                        */
int read_wrapper(int sock, SSL* ssl_token, char* buf, int count) {              //d26322 
    /* non SSL read */	
    if (!SSL_enabled)                                                           //d26322
       return read(sock, buf, count);                                           //d26322

    /* SSL-enabled flag specified - SSL_read code below */

    /* Check here *just in case* token is NULL when SSL is enabled. Return */
    /* negative and let the error handling take care of exit and cleanup   */
    if (!ssl_token)                                                             //d26322
       return -1;                                                               //d26322   

    return SSL_read(ssl_token, buf, count);                                     //d26322
}                                                                               //d26322

/* Wrapper function to use regular send function for non-SSL connections */
/* or SSL_write if SSL is enabled.                                        */
int send_wrapper(int sock, SSL* ssl_token, char* buf, int count) {              //d26322 
    /* non SSL send */	                                                        //d26322  
    if (!SSL_enabled)                                                           //d26322
       return send(sock, buf, count, 0);                                        //d26322 

    /* SSL-enabled flag specified - SSL_write code below */

    /* Check here just in case token is NULL when SSL is enabled. Return */
    /* negative and let the error handling take care of exit and cleanup */
    if (!ssl_token)                                                             //d26322
       return -1;                                                               //d26322

    return SSL_write(ssl_token, buf, count);                                    //d26322
}                                                                               //d26322

/* Helper function to perform repeated SSL_reads/reads, for longer message */ 
int read_n_loop(int sock, SSL *myssl, char *message, int message_length) {      //d26322
    int i = 0;                                                                  //d26322        
    int total = 0;                                                              //d26322
    int length = message_length;                                                //d26322
    char *message_ptr = message;                                                //d26322

    while((total < message_length) &&                                           //d26322
         ((i = read_wrapper(sock, myssl, message_ptr, length)) > 0)) {          //d26322
        total+=i;                                                               //d26322    
        length-=i;                                                              //d26322
        message_ptr+=i;                                                         //d26322
    }                                                                           //d26322

    return i;                                                                   //d26322 
}                                                                               //d26322

/* Helper function to perform repeated SSL_writes/sends for longer messages */
int send_n_loop(int sock, SSL* myssl, char* message, int message_length) {      //d26322 
    int i =0;                                                                   //d26322
    int total = 0;                                                              //d26322
    int length = message_length;                                                //d26322
                                                                      
    /* Send our response to incoming heartbeat */
    while((total < message_length) &&                                           //d26322
         ((i = send_wrapper(sock, myssl, message+total, length))>=0)) {         //d26322
         total+=i;                                                              //d26322
         length-=i;                                                             //d26322
    }                                                                           //d26322 

    return i;                                                                   //d26322
}                                                                               //d26322

/* Help screen for starting the connserver */
void conn_help() {
    printf("Program to configure a linux server and listen/respond to incoming requests.\n"
           "Intended to be used with other high speed connector socket drivers.\n"
           "Proper usage:\n"
           "./connserver -i SERVER_ADDRESS -p PORT -d RESPONSE_DELAY -s SEND_BUFFER_SIZE -r RCV_BUFFER_SIZE -q RESP_REQ -m RESP_MSG_SIZE -t TLS_ON*\n"
           "Parameter Desription:\n"
           "    Note that all parameters are optional unless specified as REQUIRED.\n\n"
           "    SERVER_ADDRESS: IPv4 address this server should open a socket on.\n"
           "                    Default value is 0.0.0.0.\n"
           "              PORT: Port that this socket should bind to.\n"
           "                    This parameter is REQUIRED.\n"
           "                    Valid values for PORT are from 1-65535.\n"
           "    RESPONSE_DELAY: Artificial time delay in ms to simulate network latency.\n"
           "                    Default value is 0. \n"
           "                    Must be an integer between 0 and 600000, inclusive. \n"
           "  SEND_BUFFER_SIZE: Size of buffer in bytes in which messages are sent.  Only used if RESP_REQ=1\n"
           "                    Default value is 8192.\n"
           "                    Must be an integer between 512 and 1048576, inclusive. \n"
           "  RCV_BUFFER_SIZE:  Size of buffer in bytes in which messages are received from the tpf_send_message API.\n"
           "                    Default value is 8192.\n"
           "                    Must be an integer between 512 and 1048576, inclusive. \n"
           "         RESP_REQ:  Defines whether or not this server will send a response when a message is received. \n"
           "                    Default value is 0.\n"
           "                    Must be either 0 or 1 -- 0 or no, 1 for yes.\n"
           "    RESP_MSG_SIZE:  The size, in bytes, of the response message sent by connserver if RESP_REQ=1.\n"
           "                    Default value is 4096.\n"
           "         * TLS_ON:  TLS_ON is NOT an associated value with option -t. Specifying -t flag toggles SSL/TLS on.\n"
           "                    Default execution without the -t flag has SSL/TLS off.\n");
}

/*Start routine for pthread_create, performs send/receive*/
/*   processing for the socket passed by threadarg       */
void *socket_handler(void *args) {
    /* Retrieve thread arguments */
    struct thread_args *actual_args = (struct thread_args *)args;               //d26322 
    int connsock = actual_args->sock_fd;  // copy socket passed                 //d26322
    SSL_CTX* ctx = actual_args->ssl_ctx;  // copy SSL context                   //d26322 
    free(actual_args);                                                          //d26322

    /* Message related parameters */
    int message_length = 0;
    unsigned short int message_type = 0;            
    char  byte_count[4];	                        //holds byte count of message given in the header
    char  hb_response_buffer[HSC_HEADER_SIZE];      //buffer containing the heartbeat response
    char* message;                                  //holding the message for read in of size byte_count
    struct hsc_resp *resp_buffer;

    /* Return codes */
    int rc = 0;                           //rc used in response message         //d26322
    int err;                                                                    //d26322

    /* SSL related parameters */
    SSL* myssl = NULL;                    // SSL token, NULL by default and stays NULL if no -s option specified 

    /* SSL related tasks if -s flag is specified */
    if (SSL_enabled) {                                                          //d26322
       /* If SSL option specified, have to do additional steps to modify the socket */
       /* Create new SSL structure from SSL context */
       myssl = SSL_new(ctx);                                                    //d26322
 
       if(!myssl) {                                                             //d26322
         printf("Error creating SSL structure.\n");                             //d26322
         exit(0);                                                               //d26322
       }                                                                        //d26322

       /* Binds the SSL object with the socket */
       SSL_set_fd(myssl, connsock);                                             //d26322 

       /* SSL Handshake and associated error checking */
       err = SSL_accept(myssl);                                                 //d26322
       if (err < 1) {                                                           //d26322
          err = SSL_get_error(myssl, err);                                      //d26322
          printf("SSL error #%d in SSL_accept\n", err);                         //d26322

          /* Error with sockets, SSL_ERROR_SYSCALL corresponds to 5 */
          if (err == SSL_ERROR_SYSCALL)                                         //d26322    
             printf("Socket error during handshake\n");                         //d26322
    
             conn_err_t("Error in SSL Handshake.\n", connsock, myssl);          //d26322
       }                                                                        //d26322

       else                                                                     //d26322
          printf("SSL handshake successful\n");                                 //d26322 
    }                                                                           //d26322

    while(1) {   //send and receive indefinitely
        /* Read in the 4 byte bytecount of incoming message. Handle if no incoming data was read */
        err = read_n_loop(connsock, myssl, byte_count, 4);                      //d26322
        if (err <= 0) {                                                         //d26322
           if (SSL_enabled) {                                                   //d26322
              // i is the return for SSL_read, which will contain error code
              err = SSL_get_error(myssl, err);                                  //d26322

              /* If error code is 6, means client issued SSL_shutdown. Respond with             */
              /* shutdown to complete graceful shutdown. Error code is is SSL_ERROR_ZERO_RETURN */
              if (err == SSL_ERROR_ZERO_RETURN)                                 //d26322 
                 SSL_shutdown(myssl);                                           //d26322
           }                                                                    //d26322

           conn_err_t("Error reading in byte count, thread exits",              //d26322
                      connsock, myssl);                                         //d26322
        }                                                                       //d26322

        // Convert from big to little Endian if server is latter 
        memcpy(&message_length, ntohl(byte_count), 4);                          //d26480
        message = malloc(message_length);

        /*Read in the rest of the incoming message. Handle if no incoming data was read */
        err = read_n_loop(connsock, myssl, message, message_length);            //d26322               
        if (err <= 0) {                                                         //d26322  
           if (SSL_enabled) {                                                   //d26322  
              // i is the return for SSL_read, which will contain the error code
              err = SSL_get_error(myssl, err);                                  //d26322

              /* If error code is 6, means client issued SSL_shutdown. Respond with             */
              /* shutdown to complete graceful shutdown. Error code is is SSL_ERROR_ZERO_RETURN */
              if (err == SSL_ERROR_ZERO_RETURN)                                 //d26322
                 SSL_shutdown(myssl);                                           //d26322
           }                                                                    //d26322
  
           free(message);                                                       //d26322
           conn_err_t("Error reading in message, thread exits",                 //d26322
                       connsock, myssl);                                        //d26322
        }

        /* Copy message type from header. */
        memcpy(&message_type, message + 2, 2);

        // Convert from big to little Endian if server is latter 
        message_type = ntohs(message_type);                                     //d26480

        /* Send response for the heartbeat */
        if(message_type == 1) {
            resp_buffer = (struct hsc_resp *) hb_response_buffer;
            
            /* Message length is equal to header - length of byte count field. */
            message_length = HSC_HEADER_SIZE - sizeof(resp_buffer->length);

            // Convert message_length to network byte order if necc
            resp_buffer->length = htonl(message_length);                        //d26480
            
            /* Fill next 2 bytes with header size. */
            resp_buffer->header_length = htons(HSC_HEADER_SIZE);                //d26480
            rc=0;
            
            /* Fill next 2 bytes with a return code. */
            resp_buffer->rc = htons(rc);                                        //d26480
            
            /* Copy header from input message to response message. */
            memcpy(resp_buffer->reserve, message+4, sizeof(resp_buffer->reserve));
            
            /* Set amount of data to send. */
            message_length = HSC_HEADER_SIZE;
            
            /* Send heartbeat. Handle if no data was able to be sent. */
            err = send_n_loop(connsock, myssl, hb_response_buffer, message_length);
            if(err<=0) {                                                        //d26322
              if (SSL_enabled) {                                                //d26322
                 err = SSL_get_error(myssl, err);                               //d26322

                 /* If error code is 6, means client issued SSL_shutdown. Respond with             */
                 /* shutdown to complete graceful shutdown. Error code is 6 SSL_ERROR_ZERO_RETURN */
                 if (err == SSL_ERROR_ZERO_RETURN)                              //d26322
                    SSL_shutdown(myssl);                                        //d26322
              }                                                                 //d26322

              free(message);                                                    //d26322
              conn_err_t("Error on send from thread", connsock, myssl);     
            }                                                                   //d26322 
        }

        else if(response_required) { // Send a response 
            char* response_buffer = malloc(response_message_size+HSC_HEADER_SIZE);
            resp_buffer = (struct hsc_resp *) response_buffer;

            /* Message length does not include byte count field. */
            message_length = response_message_size +
                             (HSC_HEADER_SIZE - sizeof(resp_buffer->length));

            // Convert message_length to network byte order if necc
            resp_buffer->length = htonl(message_length);                        //d26480
            
            /* Fill next 2 bytes with header size. */
            resp_buffer->header_length = htons(HSC_HEADER_SIZE);                //d26480
            rc=0;
            
            /* Fill next 2 bytes with return code. */
            resp_buffer->rc = htons(rc);                                        //d26480
            
            /* Copy header from input message to response message. */
            memcpy(resp_buffer->reserve, message + 4, sizeof(resp_buffer->reserve));
            
            /* If there is a response message, copy data to buffer.          */
            /* Since this is just for testing, fill response with EBCDIC B's */
            if (response_message_size > 0)
               memset(resp_buffer->data, 0xC2, response_message_size);

            /* Set number of bytes to be sent. */
            message_length = response_message_size + HSC_HEADER_SIZE;

            /* Our artifical response delay. Go to sleep, in microseconds */
            usleep(1000*response_delay);
            
            /* Send response. Handle if no data able to be sent. */
            err = send_n_loop(connsock, myssl, response_buffer, message_length);//d26322  
            if(err <= 0) {                                                      //d26322
              if (SSL_enabled) {                                                //d26322
                 err = SSL_get_error(myssl, err); // i is the return for SSL_read, which will contain the error code

                 /* If error code is 6, means client issued SSL_shutdown. Respond with             */
                 /* shutdown to complete graceful shutdown. Error code 6 is SSL_ERROR_ZERO_RETURN */
                 if (err == SSL_ERROR_ZERO_RETURN)                              //d26322
                    SSL_shutdown(myssl);                                        //d26322
              }                                                                 //d26322

              free(response_buffer);                                            //d26322  
              free(message);
              conn_err_t("Error on send from thread", connsock, myssl);         //d26322
            }

            /* free the response buffer before next iteration */
            free(response_buffer);                                              //d26322
        }
         
        /* Free our message buffer before next iteration */
        free(message);
    }
    /* Done sending messages - cleanup socket etc */
    if (SSL_enabled) {                                                          //d26322
       /* Attempt SSL shutdown */
       err = SSL_shutdown(myssl);                                               //d26322

       if (err < 0)                                                             //d26322
          printf("Error in shutdown.\n");                                       //d26322
       else if (err == 1)                                                       //d26322
          printf("Server exited gracefully.\n");                                //d26322

       SSL_free(myssl);                                                         //d26322 
       SSL_CTX_free(ctx);                                                       //d26322
    }                                                                           //d26322

    /*Really, we should never get here, but just in case. */ 
    close(connsock);                                                            //d26322
    pthread_exit(NULL);
}

/*Main entry point into program*/
int main(int argc, char *argv[]) {
    /* Initialize default IP value to 0.0.0.0 */
    inet_aton("0.0.0.0", &server_address.sin_addr);

    /* Read in option values */
    opterr = 0;
    int c;
    int optval = 1;

    /*Parse the input*/
    while ((c = getopt (argc, argv, "ti:d:s:r:q:m:p:")) != -1)                  //d26322
        switch (c)
        {
            case 'i':
                if(inet_aton(optarg, &server_address.sin_addr) == 0)
                   conn_err("Invalid IP Address");
                break;
            case 'd':
                response_delay = atoi(optarg);
                break;
            case 's':
                send_buffer_size = atoi(optarg);
                break;
            case 'r':
                receive_buffer_size = atoi(optarg);
                break;
            case 'q':
                response_required = atoi(optarg);
                break;
            case 'm':
                response_message_size = atoi(optarg);
                break;
            case 'p':
                port = atoi(optarg);
                break;
            case 't':                                                           //d26322
                SSL_enabled = 1;                                                //d26322
                break;                                                          //d26322  
            case '?':
                if (optopt == 'i' || optopt == 'd' || optopt == 's' || optopt == 'r' || optopt == 'q' || optopt == 'm')
                    printf ("CONN: Option -%c requires an argument.\n", optopt);
                else
                   conn_help();
                return 1;
            default:
                exit(0);
        }

    /*Some basic error checking on input parms, limits are specified in the documenation*/
    if(port == -1)
        conn_err("Port is a required parameter.  Re-enter the command with a valid port (option -p)"); 
    if(port<=0 || port>=65536)
        conn_err("Invalid port value.  Must fall on or between 1025-65535.");
    if(response_delay<MIN_DELAY || response_delay>MAX_DELAY)
        conn_err("Response delay is outside of acceptable range. Must fall on or between 0 and 600000");
    if(send_buffer_size<MIN_BUFFER_SIZE || send_buffer_size>MAX_BUFFER_SIZE)
        conn_err("Invalid value for send buffer size");
    if(receive_buffer_size<MIN_BUFFER_SIZE || receive_buffer_size>MAX_BUFFER_SIZE)
        conn_err("Invalid value for receive buffer size");
    if(response_required != 0 && response_required != 1)
        conn_err("Invalid value for response required - must be 0 or 1");
    if(response_message_size < 0)
        conn_err("Response message size must be greater than 0");               //d26480
   
    int sockfd, connsock;      // Parent socket, listener socket                //d26322      
    pthread_t tid;	                                                            //d26322
    SSL_CTX *ctx = NULL;       // SSL context requred to create SSL token       //d26322 

    /* Establishes SSL Context if necc. SSL Context is collection of ciphers, */
    /* trusted certficates, etc. It is set here to avoid having to recreate   */
    /* all the same settings once a connection has been established.          */
    if (SSL_enabled) {                                                          //d26322
       /* Starts up server and gets the SSL context */ 
       ctx = SSL_init_server_CTX();                                             //d26322

       /* Set Cipher List */
       if (SSL_CTX_set_cipher_list(ctx, CIPHER_LIST) <= 0) {                    //d26322
          printf("Error setting up cipher list.\n");                            //d26322
          SSL_CTX_free(ctx);                                                    //d26322
          exit(0);                                                              //d26322
       }                                                                        //d26322

       /* Set up and load certificates */
       LoadCertificates(ctx, CERT_FILE, KEY_FILE);                              //d26322  
    }

    /* Create the socket */
    sockfd = socket(AF_INET, SOCK_STREAM, 0);
    if(sockfd < 0) {                                                            //d26322
      if (SSL_enabled)                                                          //d26322
         SSL_CTX_free(ctx);                                                     //d26322

      conn_err("Error opening socket");
    }

    /* Allow a socket to be forcibly bound to a port in use by another socket */
    setsockopt(sockfd, SOL_SOCKET, SO_REUSEADDR,
               (char *)&optval, sizeof(optval));

    /* Set the address and port to values input or default values */
    server_address.sin_family = AF_INET;
    server_address.sin_port = htons(port);

    /* Bind our socket to the specified address */
    if(bind(sockfd, (struct sockaddr *) &server_address, sizeof(server_address)) < 0)
        conn_err_s("Error on binding", sockfd, ctx);                            //d26322

    /* Set the send and receive buffer.  It is important to note that the actual  *
     *    size of these buffers, which can be seen by a getsockopt() call here,   *
     *    is DOUBLE whatever the size specified in the setsockopt call is.        */
    if(setsockopt(sockfd, SOL_SOCKET, SO_RCVBUF, (char *)&receive_buffer_size, sizeof(int)) < 0)
        conn_err_s("Error on setsockopt with receive_buffer_size option", sockfd, ctx);

    if(setsockopt(sockfd, SOL_SOCKET, SO_SNDBUF, (char *)&send_buffer_size, sizeof(int)) < 0)
        conn_err_s("Error on setsockopt with send_buffer_size option", sockfd, ctx);

    /*Mark the socket as accepting connections */
    listen(sockfd, 500);

    /* Loop on accept and create a new socket for all incoming connections.        */
    while ((connsock = accept(sockfd, NULL, NULL))) {
        if (connsock <= 0)
            conn_err_s("accept() error", sockfd, ctx);                          //d26322

    /* Copy the socket into a new buffer before passing it to pthread_create.  This memory is freed in the *
     *   socket handling routine.  This is necessary because if we are accepting connections very rapidly, *
     *   it is possible that we could receive a new value for connsock from the accept function before it  *
     *   is copied to a local variable in the socket_handler routine.                                      */
        struct thread_args *args = malloc(sizeof(struct thread_args*));         //d26322

        memcpy(&(args->sock_fd), &connsock, sizeof(args->sock_fd));             //d26322
        args->ssl_ctx = ctx;    // Assignment is used rather than memcpy in case ctx is a null pointer

    /* Invoke our socket handler function which will perform reads and sends from  *
        the incoming connection in the newly created socket from accept()          */
        if(pthread_create( &tid, NULL, socket_handler, args) < 0)               //d26322
          conn_err_s("Failure creating thread", sockfd, ctx);                   //d26322
    }

    return 0;
}
