/**********************************************************************/
/*      Licensed Materials - Property of IBM                          */
/*      "Restricted Materials of IBM"                                 */
/*      5748-T15                                                      */
/*      Copyright IBM Corp. 2016, 2017                                */
/*      All Rights Reserved                                           */
/*      US Government Users Restricted Rights - Use, duplication      */
/*      or disclosure restricted by GSA ADP Schedule Contract         */
/*      with IBM Corp.                                                */
/**********************************************************************/
/*                                                                    */
/* PROGRAM NAME......QHSR                                             */
/* PURPOSE...........Read in data from High Speed Connector client.   */
/*                                                                    */
/* DESCRIPTION: QHSR reads in data from the client using the          */
/*              activate_on_receipt_with_length() function to read in */
/*              the byte count and the read() function to read in the */
/*              rest of the data. If a response if requested from the */
/*              client, QHSR sends a response back to the client.     */
/*                                                                    */
/**********************************************************************/
/*                                                                    */
/* DEPENDENCIES... None.                                              */
/*                                                                    */
/* RESTRICTIONS... None.                                              */
/*                                                                    */
/* SPECIAL CONSIDERATIONS...                                          */
/*                                                                    */
/* INTERFACES: Program QHSR is activated by an                        */
/*             activate_on_receipt_by_length() function from program  */
/*             QHSS on first message received from socket and from    */
/*             QHSR on each subsequent message received for socket.   */
/*                                                                    */
/**********************************************************************/
/*                                                                    */
/* CHANGE ACTIVITY...                                                 */
/*                                                                    */
/*   DEFECT     DATE       DESCRIPTION                                */
/*   ----       ----       -----------                                */
/*   26077      04/28/2016 Create High Speed Connector server driver  */
/*   26200      12/15/2017 Driver enhancements                        */
/**********************************************************************/

#include <stdlib.h>
#include <sys/socket.h>        //needed for sockets
#include <tpf/connapi.h>
#include <tpf/tpfapi.h>

#define PRIMECRAS 0x010000

void QHSR() {

    /* Required for reading from ECB work area */
    char      *read_addr;
    int       read_len;
    int       conn_sock;
    /* Parameters only if we are responding */
    int       response_delay;
    int       response_message_size;
    int       response_required = 1;

    /* variables used for send/read functions */
    int       rc;  
    int       message_length;
    unsigned short message_type;
    unsigned short header_size;
    char      *message;
    
    /* For use by tpf_tcpip_message_cnt*/
    char      proto[4] = "TCP"; 
    int       cnt, port, messages;
    struct    sockaddr_in conn_sock_addr; 
    struct    hsc_resp *resp_buffer;
    int       addrlen = sizeof(conn_sock_addr);    

    /* Copy info passed through the ECB work area */
    conn_sock = (int)ecbptr()->ebrout;
    ecbptr()->ebrout = PRIMECRAS;
    memcpy(&read_addr,&(ecbptr()->ebw024), sizeof(read_addr));
    memcpy(&read_len, &(ecbptr()->ebw016),sizeof(read_len));
    
    /* If no socket was passed to QHSR, program dies */
    if(!conn_sock)
        exit(0);
        
    /* No buffer to read from */
    if(read_addr == NULL)
    {
        close(conn_sock);
        exit(0);
    }
 
    /* Copy rest of data from ECB work area */
    memcpy(&response_delay, &ecbptr()->ebw004, sizeof(response_delay));
    memcpy(&response_message_size, &ecbptr()->ebw008, sizeof(response_message_size));

    /* If response message size is less than 0, no response is required. */
    if(response_message_size < 0)
        response_required = 0;

    /* Read the 4 byte bytecount from the incoming message, of the incoming message */
    rc = read(conn_sock,read_addr,read_len);
    if (rc<=0)  // Bytecount wasn't able to be read
    {
        close(conn_sock);
        exit(0);
    }
    else        // Copy bytecount to message_length buffer
    {
        memcpy(&message_length, read_addr, sizeof(message_length));
        message = malloc(message_length);
    }
    
    //Reset counters for subsequent read call
    int i = 0;
    int total = 0;
    int length = message_length;
    char *message_ptr = message;
    
    /*Read in the rest of the incoming message*/
    while((total < message_length) && ((i = read(conn_sock, message_ptr, length)) > 0))
    { 
        total+=i;  
        length = length - i; 
        message_ptr+=i; 
    }
  
    if (i <= 0) //The rest of the incoming message wasn't able to be read in
    { 
        close(conn_sock);
        exit(0);
    }
    
    // Copy the rest of the header into appropriate fields
    memcpy(&header_size, message, sizeof(header_size));
    memcpy(&message_type, message+2, sizeof(message_type));
 
    // Increment tcp/ip msg counter aftermessage is sent and we find the port
    if(getsockname(conn_sock,(struct sockaddr *)&conn_sock_addr, &addrlen)!= 0) // socket structure
        exit(0);  //getsockname failed
        
    // Get port of socket
    port = conn_sock_addr.sin_port;
    messages = 1;
    cnt = NSDB_INPUT_CNT;
    
    // If message type is not a heartbeat, increment TCP/IP message count
    if(message_type != 1)                
        tpf_tcpip_message_cnt(cnt, port, proto, messages);
 
    if(message_type == 1)  //send response for heartbeat
    {
        /* Declare/reinitialize values for send call */
        char hb_response_buffer[HSC_HEADER_SIZE];
        resp_buffer = (struct hsc_resp *) hb_response_buffer;
        i=0;
        total=0;
        
        /* Message length is equal to header - length of byte count field. */
        message_length = HSC_HEADER_SIZE - sizeof(resp_buffer->length);
        resp_buffer->length = message_length;
        
        /* Fill next 2 bytes with header length. */
        resp_buffer->header_length = HSC_HEADER_SIZE;
        rc=0;
        
        /* Fill next 2 bytes with return code. */
        resp_buffer->rc = rc;
        
        /* Copy header from input message to response message. */
        memcpy(resp_buffer->reserve, message+4, sizeof(resp_buffer->reserve));
        message_length = HSC_HEADER_SIZE;
        length = message_length;
        
        /* Send response for heartbeat */
        while((total < message_length) && ((i = send(conn_sock, hb_response_buffer+total, length, 0))>=0))
        {  
            total+=i; 
            length-=i;
        }
        /* Send was blocked, exit */
        if(i<0)
            exit(0);
	}
    else if(response_required) /* Response message was requested, send a response */
    {
        /* Create buffer/reinitialize variables for send */
        char* response_buffer = malloc(response_message_size+HSC_HEADER_SIZE);
        resp_buffer = (struct hsc_resp *) response_buffer;
        i=0;
        total=0;
        
        /* Message length does not include byte count field. */
        message_length = response_message_size +
                         (HSC_HEADER_SIZE - sizeof(resp_buffer->length));
        resp_buffer->length = message_length;
        
        /* Fill next 2 bytes with header length. */
        resp_buffer->header_length = HSC_HEADER_SIZE;
        rc=0;
        resp_buffer->rc = rc;
        
        /* Copy header from input message to response message. */
        memcpy(resp_buffer->reserve, message + 4, sizeof(resp_buffer->reserve));
        
        /* If response message needed, fill buffer with data. */
        if (response_message_size > 0)
            memset(resp_buffer->data, 0xC2, response_message_size);
            
        /* Number of bytes to be sent includes header and response message (optional). */
        message_length = HSC_HEADER_SIZE + response_message_size;
        length = message_length;

        //The maximum usleep we can issue is 1000000....otherwise we must just use sleep if greater than that.
        if (response_delay < 1000) {
            usleep(1000*response_delay);   //Artificial response delay
        }
        else {
            int secs = (response_delay / 1001) + 1;
            sleep (secs);
        }
        
        /*Send a response message to TPF*/
        while((total < message_length) && ((i = send(conn_sock, response_buffer+total, length, 0))>=0))
        {  
            total+=i; 
            length-=i;	
        }
  
        /* Data wasn't able to be sent, die */
        if(i<0)
            exit(0);
            
        // Increment tcp/ip msg counter after message is sent and we find the port
        port = conn_sock_addr.sin_port;
        messages = 1;
        cnt = NSDB_OUTPUT_CNT;
        tpf_tcpip_message_cnt(cnt, port, proto, messages);
        free(response_buffer);
    }
    
    /* This ECB finishes, aorwl qhsr again */
    unsigned char aorl_parm[8],aorl_pgm[4];
    memcpy(aorl_pgm, "QHSR", sizeof(aorl_pgm));
    memset(aorl_parm, 0, sizeof(aorl_parm));
    /* If response required, pass response delay and response message size */
    /* to program QHSR when additional data from client arrives.           */
    if(response_required)
    {
        memcpy(aorl_parm, &response_delay, sizeof(response_delay));
        memcpy(aorl_parm+4, &response_message_size, sizeof(response_message_size));
    }
    /* If response is not required, we pass -1 for response message size to */
    /* program QHSR when additional data from client arrives.               */
    else
    {
        response_message_size = -1;
        memcpy(aorl_parm+4, &response_message_size, sizeof(response_message_size));
    }
    rc = activate_on_receipt_with_length(conn_sock, aorl_parm, aorl_pgm, 4);
    /* If activate_on_receipt_with_length() returns an error,  */
    /* close the socket.                                       */ 
    if (rc == -1)
        close(conn_sock);                                                   
    exit(0);

}
