/**********************************************************************/
/*      Licensed Materials - Property of IBM                          */
/*      "Restricted Materials of IBM"                                 */
/*      5748-T15                                                      */
/*      Copyright IBM Corp. 2016, 2018                                */
/*      All Rights Reserved                                           */
/*      US Government Users Restricted Rights - Use, duplication      */
/*      or disclosure restricted by GSA ADP Schedule Contract         */
/*      with IBM Corp.                                                */
/**********************************************************************/
/*                                                                    */
/* FUNCTION NAME.....QHSS                                             */
/* PURPOSE...........QHSS driver                                      */
/* FUNCTION TYPE.....TPF                                              */
/*                                                                    */
/* INPUT.......ZINET ACT                                              */
/*                                                                    */
/* Options:     zinet add s-connXXXX port-XXXX prot-tcp pgm-qhss      */
/*              model-aoa2 state-cras act-auto                        */
/*                                                                    */
/*              s-connXXXX specifies the server name, where XXXX is a */
/*                  4-digit number that matches the server cfg file,  */
/*                  where the cfg file contains the server's parms    */
/*                  and is located in the etc/conn directory.         */
/*                                                                    */
/*              port-XXXX, where XXXX represents the port number      */
/*                                                                    */
/* DESCRIPTION: QHSS is the server program for testing the high       */
/*              speed connector.  This program parses the config file */
/*              for a TPF endpoint and then starts the QHSR program,  */
/*              which will start a persistent socket with the         */
/*              parameters specified in the config file.              */
/*                                                                    */
/**********************************************************************/
/*                                                                    */
/* DEPENDENCIES... None.                                              */
/*                                                                    */
/* RESTRICTIONS... None.                                              */
/*                                                                    */
/* SPECIAL CONSIDERATIONS...                                          */
/*                                                                    */
/* INTERFACES:                                                        */
/*                                                                    */
/*                                                                    */
/**********************************************************************/
/*                                                                    */
/* CHANGE ACTIVITY...                                                 */
/*                                                                    */
/*   DEFECT     DATE       DESCRIPTION                                */
/*   ----       ----       -----------                                */
/*   f26077     04/28/2016 Create High Speed Connector server driver  */
/*   f26200     12/15/2017 Add stdlib.h include statement and set     */
/*                         AOR_BALANCE on sockets                     */
/*   d26322     08/24/2018 Modify message prefix.                     */
/**********************************************************************/

#include <stdlib.h>
#include <sys/socket.h>        //needed for sockets
#include <sys/ioctl.h>         //needed for ioctl                               //f26200
#include <tpf/tpfapi.h>
#include <qhss.h>              //includes structure for activate on receipt with length

#define MAX_BUFFER_SIZE 1048576  //Max send/receive buffer size
#define MIN_BUFFER_SIZE 512      //Min send/receive buffer size
#define MAX_DELAY 600000         //Maximum artificial delay
#define MIN_DELAY 0              //Minimum artificial delay

#define CONNDIR         "/etc/conn"          //high speed connector directory, specifying location of config file
/* QHSS Error Messages */
#define QHSS0110TXT     "CONN ERROR, ERROR OPENNING FILE"
#define QHSS0111TXT     "CONN ERROR, SOCKET INVALID"
#define QHSS0112TXT     "CONN ERROR, Invalid response delay - must be between 0 and 600000 inclusive."
#define QHSS0113TXT     "CONN ERROR, Invalid value for send buffer size."
#define QHSS0114TXT     "CONN ERROR, Invalid value for receive buffer size."
#define QHSS0115TXT     "CONN ERROR, Invalid value for response required - must be 0 or 1."
#define QHSS0116TXT     "CONN ERROR, Invalid response message size."
#define QHSS0117TXT     "CONN ERROR, Error on setsockopt with receive_buffer_size option."
#define QHSS0118TXT     "CONN ERROR, Error on setsockopt with send_buffer_size option."
#define QHSS0119TXT     "CONN ERROR, Specified directory does not exist"

#define QHSC_MSG_PREFIX "QHSS"
#define DRVR_MSG_HEADER 1
#define API_HEADER_LEN 4

void QHSS() {

   char      message_buffer[400];             //Message buffer for printing messages

   int       conn_sock;                        //Socket for reading and sending msgs

   /*Server parameters with default values, will use these unless specified in .cfg file */
   int       response_delay = 0;              //Artificial response delay to simulate network load
   int       send_buffer_size = 8192;         //Size of the send buff
   int       receive_buffer_size = 8192;      //Size of the receive buffer
   int       response_required = 0;           //Response required from server if 1 on 
                                              //  receiving of a message
   int       response_message_size = 4096;    //Size of the response message
   /********************/
   int       count=0;                         //for file parsing
   char      parm[5][32];
   int       rc;
   
   //Retrieve socket passed from aoa2
   (void)memcpy(&conn_sock,&(ecbptr()->ebw008),sizeof(conn_sock));  
   if(conn_sock<=0)
   {
           sprintf(message_buffer,QHSS0111TXT);
           conn_drvr_message(message_buffer,
                             WTOPC_NO_CHAIN,
                             DRVR_MSG_HEADER,
                             111,
                             'E',
                             QHSC_MSG_PREFIX);                                  //d26322
           exit(0);
   }   

   /*******************************************************/
   /* Set AOR Balance to balance requests across cpus     */
   /*******************************************************/
   int arg = 1;                                                                 //f26200
   rc = ioctl(conn_sock, AOR_BALANCE, (char *) &arg);                           //f26200
   if(rc!=0)  //return error                                                    //f26200
   {                                                                            //f26200
       printf("error on call to ioctl\n");                                      //f26200
       rc = close(conn_sock);                                                   //f26200
       exit(0);                                                                 //f26200
   }                                                                            //f26200
   
   /*******************************************************/
   /* Call getsockname to obtain the server's port number */
   /*******************************************************/
   struct sockaddr_in name_sockaddr_in;
   int sockaddr_in_length = sizeof(name_sockaddr_in);
   
   rc = getsockname(conn_sock, (struct sockaddr *)&name_sockaddr_in, 
                    &sockaddr_in_length);
   if(rc!=0)  //return error
   {
       printf("error on call to getsockname\n");
       rc = close(conn_sock);
       exit(0);
   }
   /*******************************************************/
   /* Save the name of the configuration file using the   */
   /* port number                                         */
   /*******************************************************/   
   char conn_conf_file[16];   //name of input config file
   if(sprintf(conn_conf_file, "config%d.cfg", ntohs(name_sockaddr_in.sin_port)) < 0) //Insert port number into config file name
   {
       printf("sprintf failed\n");
   }
   
   /*******************************************************/
   /* Switch to /etc/conn directory and open the          */
   /* configuration file                                  */
   /*******************************************************/  
   
   if(chdir(CONNDIR) != 0)  //switch to /etc/conn directory
   {
       //directory not found, print error message
       sprintf(message_buffer,QHSS0119TXT);
       conn_drvr_message(message_buffer,
                         WTOPC_NO_CHAIN,
                         DRVR_MSG_HEADER,
                         110,
                         'E',
                         QHSC_MSG_PREFIX);                                      //d26322
       exit(0);    
   }  

   FILE *fp;
   fp = fopen(conn_conf_file, "r");
   if(fp == NULL)   //Check if file open successfully
   {
           sprintf(message_buffer,QHSS0110TXT);
           conn_drvr_message(message_buffer,
                             WTOPC_NO_CHAIN,
                             DRVR_MSG_HEADER,
                             110,
                             'E',
                             QHSC_MSG_PREFIX);                                  //d26322
           exit(0);
   }

   memset(parm,0,sizeof(parm));       //Scan input file
   rc = fscanf(fp,"%s",parm[count]);

   while( rc != EOF && count < 4){    
      count++;
      rc = fscanf(fp,"%s",parm[count]);
   }

   count=0;
   /* Parse the config file */
   while(parm[count][0]!='\0'){
      if((strncmp(parm[count],"DELAY=",6) == 0)){  // Parse DELAY field from config file

         response_delay = atoi(parm[count]+6);
         if(response_delay<MIN_DELAY || response_delay>MAX_DELAY)  //Error if delay is not in valid range
         {
             sprintf(message_buffer,QHSS0112TXT);
             conn_drvr_message(message_buffer,
                               WTOPC_NO_CHAIN,
                               DRVR_MSG_HEADER,
                               112,
                               'E',
                               QHSC_MSG_PREFIX);                                //d26322
             exit(0);
         }

      }else if((strncmp(parm[count],"SNDBUFSZ=",9) == 0)){  // Parse SNDBUFSZ field from config file

         send_buffer_size = atoi(parm[count]+9);
         if(send_buffer_size<MIN_BUFFER_SIZE || send_buffer_size>MAX_BUFFER_SIZE)  //Error if send buffer is not in valid range
         {         
             sprintf(message_buffer,QHSS0113TXT);
             conn_drvr_message(message_buffer,
                               WTOPC_NO_CHAIN,
                               DRVR_MSG_HEADER,
                               113,
                               'E',
                               QHSC_MSG_PREFIX);                                //d26322
             exit(0);
         }

      }else if((strncmp(parm[count],"RCVBUFSZ=",9) == 0)){  // Parse 'RCVBUFSZ' field from config file

         receive_buffer_size = atoi(parm[count]+9);
         if(receive_buffer_size<MIN_BUFFER_SIZE || receive_buffer_size>MAX_BUFFER_SIZE) //Error if response buffer is not in valid range
         {
             sprintf(message_buffer,QHSS0114TXT);
             conn_drvr_message(message_buffer,
                               WTOPC_NO_CHAIN,
                               DRVR_MSG_HEADER,
                               114,
                               'E',
                               QHSC_MSG_PREFIX);                                //d26322
             exit(0);
         }

      }else if((strncmp(parm[count],"RESPREQ=",8) == 0)){  // Parse 'RESPREQ' field from config file

         response_required = atoi(parm[count]+8);
         if(response_required != 0 && response_required != 1) //error if response required is not 0 or 1
         {
             sprintf(message_buffer,QHSS0115TXT); 
             conn_drvr_message(message_buffer,
                               WTOPC_NO_CHAIN,
                               DRVR_MSG_HEADER,
                               115,
                               'E',
                               QHSC_MSG_PREFIX);                                //d26322
             exit(0);
         }

      }else if((strncmp(parm[count],"RESPMSGSZ=",10) == 0)){  // Parse RESPMSGSZ field from config file

         response_message_size = atoi(parm[count]+10);
         if(response_message_size < 0)  //error if response message size is negative
         {
             sprintf(message_buffer,QHSS0116TXT);
             conn_drvr_message(message_buffer,
                               WTOPC_NO_CHAIN,
                               DRVR_MSG_HEADER,
                               116,
                               'E',
                               QHSC_MSG_PREFIX);                                //d26322
             exit(0);
         }

      }
      count++;
   }
   
 
    /*Set the send and receive buffer.  It is important to note that the actual  *
     *    size of these buffers, which can be seen by a getsockopt() call here,  *
     *    is DOUBLE whatever the size specified in the setsockopt call is.       */
    if(setsockopt(conn_sock, SOL_SOCKET, SO_RCVBUF, (char *)&receive_buffer_size, sizeof(int)) < 0){
        sprintf(message_buffer,QHSS0117TXT);
        conn_drvr_message(message_buffer,
                          WTOPC_NO_CHAIN,
                          DRVR_MSG_HEADER,
                          117,
                          'E',
                          QHSC_MSG_PREFIX);                                     //d26322
        exit(0);
    }
    if(setsockopt(conn_sock, SOL_SOCKET, SO_SNDBUF, (char *)&send_buffer_size, sizeof(int)) < 0){
        sprintf(message_buffer,QHSS0118TXT);
        conn_drvr_message(message_buffer,
                          WTOPC_NO_CHAIN,
                          DRVR_MSG_HEADER,
                          118,
                          'E',
                          QHSC_MSG_PREFIX);                                     //d26322
        exit(0);
    }
    
    // Kick off QHSR program that actually does the reads and sends
    unsigned char aorl_parm[8],aorl_pgm[4];
    memcpy(aorl_pgm, "QHSR", sizeof(aorl_pgm));
    memset(aorl_parm, 0, sizeof(aorl_parm));
    /* If response required, pass response delay and response message size */
    /* to program QHSR when data from client arrives.                      */
    if(response_required)
    {
        memcpy(aorl_parm, &response_delay, sizeof(response_delay));
        memcpy(aorl_parm+4, &response_message_size, sizeof(response_message_size));
    }
    /* If response is not required, we pass -1 for response message size to */
    /* program QHSR when data from client arrives.                          */
    else
    {
        response_message_size = -1;
        memcpy(aorl_parm+4, &response_message_size, sizeof(response_message_size));
    }
    rc = activate_on_receipt_with_length(conn_sock, aorl_parm, aorl_pgm, 4);
    /* If activate_on_receipt_with_length() returns an error, */
    /* close the socket.                                      */
    if (rc == -1)
        close(conn_sock);                                                   
    exit(0);
}
