/**********************************************************************/
/*      Licensed Materials - Property of IBM                          */
/*      "Restricted Materials of IBM"                                 */
/*      5748-T15                                                      */
/*      Copyright IBM Corp. 2018                                      */
/*      All Rights Reserved                                           */
/*      US Government Users Restricted Rights - Use, duplication      */
/*      or disclosure restricted by GSA ADP Schedule Contract         */
/*      with IBM Corp.                                                */
/**********************************************************************/
/*                                                                    */
/* PROGRAM NAME......QHST                                             */
/* PURPOSE...........Read in data from High Speed Connector client.   */
/*                                                                    */
/* DESCRIPTION:                                                       */
/*                                                                    */
/**********************************************************************/
/*                                                                    */
/* DEPENDENCIES... None.                                              */
/*                                                                    */
/* RESTRICTIONS... None.                                              */
/*                                                                    */
/* SPECIAL CONSIDERATIONS...                                          */
/*                                                                    */
/* INTERFACES: Program QHST is activated by a SSL_aor function from   */
/*             QHSU on first message received from socket and from    */
/*             QHST on each subsequent message received for socket.   */
/*                                                                    */
/**********************************************************************/
/*                                                                    */
/* CHANGE ACTIVITY...                                                 */
/*                                                                    */
/*   DEFECT     DATE       DESCRIPTION                                */
/*   ----       ----       -----------                                */
/*   26322      08/24/2018 Create High Speed Connector server driver  */
/**********************************************************************/

#include <stdlib.h>
#include <sys/socket.h>        //needed for sockets
#include <tpf/connapi.h>
#include <tpf/tpfapi.h>
#include <openssl/ssl.h>
#include <qhss.h>

#define PRIMECRAS 0x010000

/* Helper function to perform repeated SSL_reads for messages of uncertain length*/ 
int SSL_read_n_loop (SSL *myssl, char *message, int message_length) {
	int i = 0;
	int total = 0;
	int length = message_length;
	char *message_ptr = message;

    while((total < message_length) && ((i = SSL_read(myssl, message_ptr, length)) > 0)) { 
        total+=i;  
        length-=i; 
        message_ptr+=i; 
    }

	return i;
}

void QHST() {

    /* Required for reading from ECB work area */
    int       conn_sock;
	SSL       *myssl = NULL;

    /* Parameters only if we are responding */
    int       response_delay, response_message_size;
    int       response_required = 1;

    /* variables used for send/read functions */
    int       rc, message_length;
    unsigned short message_type, header_size;
    char      *message;
    
    /* For use by tpf_tcpip_message_cnt*/
    char      proto[4] = "TCP"; 
    int       cnt, port, messages;
    struct    sockaddr_in conn_sock_addr; 
    struct    hsc_resp *resp_buffer;
    int       addrlen = sizeof(conn_sock_addr);    

    /* Copy info passed through the ECB work area */
    ecbptr()->ebrout = PRIMECRAS;
    memcpy(&myssl,&(ecbptr()->ebw024), sizeof(myssl));
    memcpy(&conn_sock, &(ecbptr()->ebw016),sizeof(conn_sock));
    
    /* If no socket was passed to QHST, program dies */
    if(!conn_sock)
        exit(0);
        
    /* Copy rest of data from ECB work area */
    memcpy(&response_delay, &ecbptr()->ebw004, sizeof(response_delay));
    memcpy(&response_message_size, &ecbptr()->ebw008, sizeof(response_message_size));

    /* If response message size is less than 0, no response is required. */
    if(response_message_size < 0)
        response_required = 0;

    /* Read the 4 byte bytecount from the incoming message */
	char byte_count[4];
    	
    if (SSL_read_n_loop(myssl, byte_count, 4)<=0)  // Bytecount wasn't able to be read
    {
        SSL_cleanup(myssl, conn_sock);
        exit(0);
    }

    else        // Copy bytecount to message_length buffer
    {
        memcpy(&message_length, byte_count, sizeof(message_length));
        message = malloc(message_length);
    }
    
    //Reset counters for subsequent read call
    int length = message_length;
    
    /*Read in the rest of the incoming message*/
    if (SSL_read_n_loop(myssl, message, length) <= 0) //The rest of the incoming message wasn't able to be read in
    { 
		SSL_cleanup(myssl, conn_sock);
        exit(0);
    }
    
    // Copy the rest of the header into appropriate fields
    memcpy(&header_size, message, sizeof(header_size));
    memcpy(&message_type, message+2, sizeof(message_type));
 
    // Increment tcp/ip msg counter aftermessage is sent and we find the port
    if(getsockname(conn_sock,(struct sockaddr *)&conn_sock_addr, &addrlen)!= 0) // socket structure
	{
		free(message);
		SSL_cleanup(myssl, conn_sock);
        exit(0);  //getsockname failed
	}
        
    // Get port of socket
    port = conn_sock_addr.sin_port;
    messages = 1;
	int i = 0;
	int total = 0;
    cnt = NSDB_INPUT_CNT;
    
    // If message type is not a heartbeat, increment TCP/IP message count
    if(message_type != 1)                
        tpf_tcpip_message_cnt(cnt, port, proto, messages);
 
    if(message_type == 1)  //send response for heartbeat
    {
        /* Declare/reinitialize values for send call */
        char hb_response_buffer[HSC_HEADER_SIZE];
        resp_buffer = (struct hsc_resp *) hb_response_buffer;
        i=0;
        total=0;
        
        /* Message length is equal to header - length of byte count field. */
        message_length = HSC_HEADER_SIZE - sizeof(resp_buffer->length);
        resp_buffer->length = message_length;
        
        /* Fill next 2 bytes with header length. */
        resp_buffer->header_length = HSC_HEADER_SIZE;
        rc=0;
        
        /* Fill next 2 bytes with return code. */
        resp_buffer->rc = rc;
        
        /* Copy header from input message to response message. */
        memcpy(resp_buffer->reserve, message+4, sizeof(resp_buffer->reserve));
        message_length = HSC_HEADER_SIZE;
        length = message_length;
        
        /* Send response for heartbeat */
        while((total < message_length) && ((i = SSL_write(myssl, hb_response_buffer+total, length))>=0))
        {  
            total+=i; 
            length-=i;
        }
        /* Send was blocked, exit */
        if(i<=0) {
			free(message);
			SSL_cleanup(myssl, conn_sock);
            exit(0);
		}
	}
    else if(response_required) /* Response message was requested, send a response */
    {
        /* Create buffer/reinitialize variables for send */
        char* response_buffer = malloc(response_message_size+HSC_HEADER_SIZE);
        resp_buffer = (struct hsc_resp *) response_buffer;
        i=0;
        total=0;
        
        /* Message length does not include byte count field. */
        message_length = response_message_size +
                         (HSC_HEADER_SIZE - sizeof(resp_buffer->length));
        resp_buffer->length = message_length;
        
        /* Fill next 2 bytes with header length. */
        resp_buffer->header_length = HSC_HEADER_SIZE;
        rc=0;
        resp_buffer->rc = rc;
        
        /* Copy header from input message to response message. */
        memcpy(resp_buffer->reserve, message + 4, sizeof(resp_buffer->reserve));
        
        /* If response message needed, fill buffer with data. */
        if (response_message_size > 0)
            memset(resp_buffer->data, 0xC2, response_message_size);
            
        /* Number of bytes to be sent includes header and response message (optional). */
        message_length = HSC_HEADER_SIZE + response_message_size;
        length = message_length;

        //The maximum usleep we can issue is 1000000....otherwise we must just use sleep if greater than that.
        if (response_delay < 1000) {
            usleep(1000*response_delay);   //Artificial response delay
        }
        else {
            int secs = (response_delay / 1001) + 1;
            sleep (secs);
        }
        
        /*Send a response message to TPF*/
        while((total < message_length) && ((i = SSL_write(myssl, response_buffer+total, length))>=0))
        {  
            total+=i; 
            length-=i;	
        }
  
        /* Data wasn't able to be sent, die */
        if(i<=0)
		{
			free(message);
			free(response_buffer);
			SSL_cleanup(myssl,conn_sock); 
            exit(0);
		}
            
        // Increment tcp/ip msg counter after message is sent and we find the port
        port = conn_sock_addr.sin_port;
        messages = 1;
        cnt = NSDB_OUTPUT_CNT;
        tpf_tcpip_message_cnt(cnt, port, proto, messages);
		free(message);
        free(response_buffer);
    }
    
    /* This ECB finishes, sslaor qhst again */
    unsigned char aor_parm[8],aor_pgm[4];
    memcpy(aor_pgm, "QHST", sizeof(aor_pgm));
    memset(aor_parm, 0, sizeof(aor_parm));
    /* If response required, pass response delay and response message size */
    /* to program QHSR when additional data from client arrives.           */
    if(response_required)
    {
        memcpy(aor_parm, &response_delay, sizeof(response_delay));
        memcpy(aor_parm+4, &response_message_size, sizeof(response_message_size));
    }
    /* If response is not required, we pass -1 for response message size to */
    /* program QHST when additional data from client arrives.               */
    else
    {
        response_message_size = -1;
        memcpy(aor_parm+4, &response_message_size, sizeof(response_message_size));
    }
    //rc = activate_on_receipt_with_length(conn_sock, aorl_parm, aorl_pgm, 4);
	rc = SSL_aor(myssl, aor_parm, aor_pgm, 0);
    /* If activate_on_receipt_with_length() returns an error,  */
    /* close the socket.                                       */ 
    if (rc == 0)
        close(conn_sock);                                                   
    exit(0);

}
